/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the "Elastic License
 * 2.0", the "GNU Affero General Public License v3.0 only", and the "Server Side
 * Public License v 1"; you may not use this file except in compliance with, at
 * your election, the "Elastic License 2.0", the "GNU Affero General Public
 * License v3.0 only", or the "Server Side Public License, v 1".
 */

package org.elasticsearch;

import org.elasticsearch.core.Assertions;
import org.elasticsearch.core.UpdateForV9;

import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.function.IntFunction;

/**
 * <p>Transport version is used to coordinate compatible wire protocol communication between nodes, at a fine-grained level.  This replaces
 * and supersedes the old Version constants.</p>
 *
 * <p>Before adding a new version constant, please read the block comment at the end of the list of constants.</p>
 */
public class TransportVersions {

    /*
     * NOTE: IntelliJ lies!
     * This map is used during class construction, referenced by the registerTransportVersion method.
     * When all the transport version constants have been registered, the map is cleared & never touched again.
     */
    static TreeSet<Integer> IDS = new TreeSet<>();

    static TransportVersion def(int id) {
        if (IDS == null) throw new IllegalStateException("The IDS map needs to be present to call this method");

        if (IDS.add(id) == false) {
            throw new IllegalArgumentException("Version id " + id + " defined twice");
        }
        if (id < IDS.last()) {
            throw new IllegalArgumentException("Version id " + id + " is not defined in the right location. Keep constants sorted");
        }
        return new TransportVersion(id);
    }

    @UpdateForV9 // remove the transport versions with which v9 will not need to interact
    public static final TransportVersion ZERO = def(0);
    public static final TransportVersion V_7_0_0 = def(7_00_00_99);
    public static final TransportVersion V_7_0_1 = def(7_00_01_99);
    public static final TransportVersion V_7_1_0 = def(7_01_00_99);
    public static final TransportVersion V_7_2_0 = def(7_02_00_99);
    public static final TransportVersion V_7_2_1 = def(7_02_01_99);
    public static final TransportVersion V_7_3_0 = def(7_03_00_99);
    public static final TransportVersion V_7_3_2 = def(7_03_02_99);
    public static final TransportVersion V_7_4_0 = def(7_04_00_99);
    public static final TransportVersion V_7_5_0 = def(7_05_00_99);
    public static final TransportVersion V_7_6_0 = def(7_06_00_99);
    public static final TransportVersion V_7_7_0 = def(7_07_00_99);
    public static final TransportVersion V_7_8_0 = def(7_08_00_99);
    public static final TransportVersion V_7_8_1 = def(7_08_01_99);
    public static final TransportVersion V_7_9_0 = def(7_09_00_99);
    public static final TransportVersion V_7_10_0 = def(7_10_00_99);
    public static final TransportVersion V_7_10_1 = def(7_10_01_99);
    public static final TransportVersion V_7_11_0 = def(7_11_00_99);
    public static final TransportVersion V_7_12_0 = def(7_12_00_99);
    public static final TransportVersion V_7_13_0 = def(7_13_00_99);
    public static final TransportVersion V_7_14_0 = def(7_14_00_99);
    public static final TransportVersion V_7_15_0 = def(7_15_00_99);
    public static final TransportVersion V_7_15_1 = def(7_15_01_99);
    public static final TransportVersion V_7_16_0 = def(7_16_00_99);
    public static final TransportVersion V_7_17_0 = def(7_17_00_99);
    public static final TransportVersion V_7_17_1 = def(7_17_01_99);
    public static final TransportVersion V_7_17_8 = def(7_17_08_99);
    public static final TransportVersion V_8_0_0 = def(8_00_00_99);
    public static final TransportVersion V_8_1_0 = def(8_01_00_99);
    public static final TransportVersion V_8_2_0 = def(8_02_00_99);
    public static final TransportVersion V_8_3_0 = def(8_03_00_99);
    public static final TransportVersion V_8_4_0 = def(8_04_00_99);
    public static final TransportVersion V_8_5_0 = def(8_05_00_99);
    public static final TransportVersion V_8_6_0 = def(8_06_00_99);
    public static final TransportVersion V_8_6_1 = def(8_06_01_99);
    public static final TransportVersion V_8_7_0 = def(8_07_00_99);
    public static final TransportVersion V_8_7_1 = def(8_07_01_99);
    public static final TransportVersion V_8_8_0 = def(8_08_00_99);
    public static final TransportVersion V_8_8_1 = def(8_08_01_99);
    /*
     * READ THE COMMENT BELOW THIS BLOCK OF DECLARATIONS BEFORE ADDING NEW TRANSPORT VERSIONS
     * Detached transport versions added below here.
     */
    public static final TransportVersion V_8_9_X = def(8_500_020);
    public static final TransportVersion V_8_10_X = def(8_500_061);
    public static final TransportVersion V_8_11_X = def(8_512_0_01);
    public static final TransportVersion V_8_12_0 = def(8_560_0_00);
    public static final TransportVersion V_8_12_1 = def(8_560_0_01);
    public static final TransportVersion V_8_13_0 = def(8_595_0_00);
    public static final TransportVersion V_8_13_4 = def(8_595_0_01);
    public static final TransportVersion V_8_14_0 = def(8_636_0_01);
    // 8.15.0+
    public static final TransportVersion WATERMARK_THRESHOLDS_STATS = def(8_637_0_00);
    public static final TransportVersion ENRICH_CACHE_ADDITIONAL_STATS = def(8_638_0_00);
    public static final TransportVersion ML_INFERENCE_RATE_LIMIT_SETTINGS_ADDED = def(8_639_0_00);
    public static final TransportVersion ML_TRAINED_MODEL_CACHE_METADATA_ADDED = def(8_640_0_00);
    public static final TransportVersion TOP_LEVEL_KNN_SUPPORT_QUERY_NAME = def(8_641_0_00);
    public static final TransportVersion INDEX_SEGMENTS_VECTOR_FORMATS = def(8_642_0_00);
    public static final TransportVersion ADD_RESOURCE_ALREADY_UPLOADED_EXCEPTION = def(8_643_0_00);
    public static final TransportVersion ESQL_MV_ORDERING_SORTED_ASCENDING = def(8_644_0_00);
    public static final TransportVersion ESQL_PAGE_MAPPING_TO_ITERATOR = def(8_645_0_00);
    public static final TransportVersion BINARY_PIT_ID = def(8_646_0_00);
    public static final TransportVersion SECURITY_ROLE_MAPPINGS_IN_CLUSTER_STATE = def(8_647_0_00);
    public static final TransportVersion ESQL_REQUEST_TABLES = def(8_648_0_00);
    public static final TransportVersion ROLE_REMOTE_CLUSTER_PRIVS = def(8_649_0_00);
    public static final TransportVersion NO_GLOBAL_RETENTION_FOR_SYSTEM_DATA_STREAMS = def(8_650_0_00);
    public static final TransportVersion SHUTDOWN_REQUEST_TIMEOUTS_FIX = def(8_651_0_00);
    public static final TransportVersion INDEXING_PRESSURE_REQUEST_REJECTIONS_COUNT = def(8_652_0_00);
    public static final TransportVersion ROLLUP_USAGE = def(8_653_0_00);
    public static final TransportVersion SECURITY_ROLE_DESCRIPTION = def(8_654_0_00);
    public static final TransportVersion ML_INFERENCE_AZURE_OPENAI_COMPLETIONS = def(8_655_0_00);
    public static final TransportVersion JOIN_STATUS_AGE_SERIALIZATION = def(8_656_0_00);
    public static final TransportVersion ML_RERANK_DOC_OPTIONAL = def(8_657_0_00);
    public static final TransportVersion FAILURE_STORE_FIELD_PARITY = def(8_658_0_00);
    public static final TransportVersion ML_INFERENCE_AZURE_AI_STUDIO = def(8_659_0_00);
    public static final TransportVersion ML_INFERENCE_COHERE_COMPLETION_ADDED = def(8_660_0_00);
    public static final TransportVersion ESQL_REMOVE_ES_SOURCE_OPTIONS = def(8_661_0_00);
    public static final TransportVersion NODE_STATS_INGEST_BYTES = def(8_662_0_00);
    public static final TransportVersion SEMANTIC_QUERY = def(8_663_0_00);
    public static final TransportVersion GET_AUTOSCALING_CAPACITY_UNUSED_TIMEOUT = def(8_664_0_00);
    public static final TransportVersion SIMULATE_VALIDATES_MAPPINGS = def(8_665_0_00);
    public static final TransportVersion RULE_QUERY_RENAME = def(8_666_0_00);
    public static final TransportVersion SPARSE_VECTOR_QUERY_ADDED = def(8_667_0_00);
    public static final TransportVersion ESQL_ADD_INDEX_MODE_TO_SOURCE = def(8_668_0_00);
    public static final TransportVersion GET_SHUTDOWN_STATUS_TIMEOUT = def(8_669_0_00);
    public static final TransportVersion FAILURE_STORE_TELEMETRY = def(8_670_0_00);
    public static final TransportVersion ADD_METADATA_FLATTENED_TO_ROLES = def(8_671_0_00);
    public static final TransportVersion ML_INFERENCE_GOOGLE_AI_STUDIO_COMPLETION_ADDED = def(8_672_0_00);
    public static final TransportVersion WATCHER_REQUEST_TIMEOUTS = def(8_673_0_00);
    public static final TransportVersion ML_INFERENCE_ENHANCE_DELETE_ENDPOINT = def(8_674_0_00);
    public static final TransportVersion ML_INFERENCE_GOOGLE_AI_STUDIO_EMBEDDINGS_ADDED = def(8_675_0_00);
    public static final TransportVersion ADD_MISTRAL_EMBEDDINGS_INFERENCE = def(8_676_0_00);
    public static final TransportVersion ML_CHUNK_INFERENCE_OPTION = def(8_677_0_00);
    public static final TransportVersion RANK_FEATURE_PHASE_ADDED = def(8_678_0_00);
    public static final TransportVersion RANK_DOC_IN_SHARD_FETCH_REQUEST = def(8_679_0_00);
    public static final TransportVersion SECURITY_SETTINGS_REQUEST_TIMEOUTS = def(8_680_0_00);
    public static final TransportVersion QUERY_RULE_CRUD_API_PUT = def(8_681_0_00);
    public static final TransportVersion DROP_UNUSED_NODES_REQUESTS = def(8_682_0_00);
    public static final TransportVersion QUERY_RULE_CRUD_API_GET_DELETE = def(8_683_0_00);
    public static final TransportVersion MORE_LIGHTER_NODES_REQUESTS = def(8_684_0_00);
    public static final TransportVersion DROP_UNUSED_NODES_IDS = def(8_685_0_00);
    public static final TransportVersion DELETE_SNAPSHOTS_ASYNC_ADDED = def(8_686_0_00);
    public static final TransportVersion VERSION_SUPPORTING_SPARSE_VECTOR_STATS = def(8_687_0_00);
    public static final TransportVersion ML_AD_OUTPUT_MEMORY_ALLOCATOR_FIELD = def(8_688_0_00);
    public static final TransportVersion FAILURE_STORE_LAZY_CREATION = def(8_689_0_00);
    public static final TransportVersion SNAPSHOT_REQUEST_TIMEOUTS = def(8_690_0_00);
    public static final TransportVersion INDEX_METADATA_MAPPINGS_UPDATED_VERSION = def(8_691_0_00);
    public static final TransportVersion ML_INFERENCE_ELAND_SETTINGS_ADDED = def(8_692_0_00);
    public static final TransportVersion ML_ANTHROPIC_INTEGRATION_ADDED = def(8_693_0_00);
    public static final TransportVersion ML_INFERENCE_GOOGLE_VERTEX_AI_EMBEDDINGS_ADDED = def(8_694_0_00);
    public static final TransportVersion EVENT_INGESTED_RANGE_IN_CLUSTER_STATE = def(8_695_0_00);
    public static final TransportVersion ESQL_ADD_AGGREGATE_TYPE = def(8_696_0_00);
    public static final TransportVersion SECURITY_MIGRATIONS_MIGRATION_NEEDED_ADDED = def(8_697_0_00);
    public static final TransportVersion K_FOR_KNN_QUERY_ADDED = def(8_698_0_00);
    public static final TransportVersion TEXT_SIMILARITY_RERANKER_RETRIEVER = def(8_699_0_00);
    public static final TransportVersion ML_INFERENCE_GOOGLE_VERTEX_AI_RERANKING_ADDED = def(8_700_0_00);
    public static final TransportVersion VERSIONED_MASTER_NODE_REQUESTS = def(8_701_0_00);
    public static final TransportVersion ML_INFERENCE_AMAZON_BEDROCK_ADDED = def(8_702_0_00);
    public static final TransportVersion ENTERPRISE_GEOIP_DOWNLOADER_BACKPORT_8_15 = def(8_702_0_01);
    public static final TransportVersion FIX_VECTOR_SIMILARITY_INNER_HITS_BACKPORT_8_15 = def(8_702_0_02);
    /**
     * we made a single backport for ESQL_ES_FIELD_CACHED_SERIALIZATION and ESQL_ATTRIBUTE_CACHED_SERIALIZATION
     * with only one TransportVersion entry
     */
    public static final TransportVersion ESQL_ATTRIBUTE_CACHED_SERIALIZATION_8_15 = def(8_702_0_03);
    public static final TransportVersion QUERY_RULES_LIST_INCLUDES_TYPES_BACKPORT_8_15 = def(8_702_0_04);
    public static final TransportVersion ML_INFERENCE_DONT_DELETE_WHEN_SEMANTIC_TEXT_EXISTS = def(8_703_0_00);
    public static final TransportVersion INFERENCE_ADAPTIVE_ALLOCATIONS = def(8_704_0_00);
    public static final TransportVersion INDEX_REQUEST_UPDATE_BY_SCRIPT_ORIGIN = def(8_705_0_00);
    public static final TransportVersion ML_INFERENCE_COHERE_UNUSED_RERANK_SETTINGS_REMOVED = def(8_706_0_00);
    public static final TransportVersion ENRICH_CACHE_STATS_SIZE_ADDED = def(8_707_0_00);
    public static final TransportVersion ENTERPRISE_GEOIP_DOWNLOADER = def(8_708_0_00);
    public static final TransportVersion NODES_STATS_ENUM_SET = def(8_709_0_00);
    public static final TransportVersion MASTER_NODE_METRICS = def(8_710_0_00);
    public static final TransportVersion SEGMENT_LEVEL_FIELDS_STATS = def(8_711_0_00);
    public static final TransportVersion ML_ADD_DETECTION_RULE_PARAMS = def(8_712_0_00);
    public static final TransportVersion FIX_VECTOR_SIMILARITY_INNER_HITS = def(8_713_0_00);
    public static final TransportVersion INDEX_REQUEST_UPDATE_BY_DOC_ORIGIN = def(8_714_0_00);
    public static final TransportVersion ESQL_ATTRIBUTE_CACHED_SERIALIZATION = def(8_715_0_00);
    public static final TransportVersion REGISTER_SLM_STATS = def(8_716_0_00);
    public static final TransportVersion ESQL_NESTED_UNSUPPORTED = def(8_717_0_00);
    public static final TransportVersion ESQL_SINGLE_VALUE_QUERY_SOURCE = def(8_718_0_00);
    public static final TransportVersion ESQL_ORIGINAL_INDICES = def(8_719_0_00);
    public static final TransportVersion ML_INFERENCE_EIS_INTEGRATION_ADDED = def(8_720_0_00);
    public static final TransportVersion INGEST_PIPELINE_EXCEPTION_ADDED = def(8_721_0_00);
    public static final TransportVersion ZDT_NANOS_SUPPORT_BROKEN = def(8_722_0_00);
    public static final TransportVersion REMOVE_GLOBAL_RETENTION_FROM_TEMPLATES = def(8_723_0_00);
    public static final TransportVersion RANDOM_RERANKER_RETRIEVER = def(8_724_0_00);
    public static final TransportVersion ESQL_PROFILE_SLEEPS = def(8_725_0_00);
    public static final TransportVersion ZDT_NANOS_SUPPORT = def(8_726_0_00);
    public static final TransportVersion LTR_SERVERLESS_RELEASE = def(8_727_0_00);
    public static final TransportVersion ALLOW_PARTIAL_SEARCH_RESULTS_IN_PIT = def(8_728_0_00);
    public static final TransportVersion RANK_DOCS_RETRIEVER = def(8_729_0_00);
    public static final TransportVersion ESQL_ES_FIELD_CACHED_SERIALIZATION = def(8_730_0_00);
    public static final TransportVersion ADD_MANAGE_ROLES_PRIVILEGE = def(8_731_0_00);
    public static final TransportVersion REPOSITORIES_TELEMETRY = def(8_732_0_00);
    public static final TransportVersion ML_INFERENCE_ALIBABACLOUD_SEARCH_ADDED = def(8_733_0_00);
    public static final TransportVersion FIELD_CAPS_RESPONSE_INDEX_MODE = def(8_734_0_00);
    public static final TransportVersion GET_DATA_STREAMS_VERBOSE = def(8_735_0_00);
    public static final TransportVersion ESQL_ADD_INDEX_MODE_CONCRETE_INDICES = def(8_736_0_00);
    public static final TransportVersion UNASSIGNED_PRIMARY_COUNT_ON_CLUSTER_HEALTH = def(8_737_0_00);
    public static final TransportVersion ESQL_AGGREGATE_EXEC_TRACKS_INTERMEDIATE_ATTRS = def(8_738_0_00);
    public static final TransportVersion CCS_TELEMETRY_STATS = def(8_739_0_00);
    public static final TransportVersion GLOBAL_RETENTION_TELEMETRY = def(8_740_0_00);
    public static final TransportVersion ROUTING_TABLE_VERSION_REMOVED = def(8_741_0_00);
    public static final TransportVersion ML_SCHEDULED_EVENT_TIME_SHIFT_CONFIGURATION = def(8_742_0_00);
    public static final TransportVersion SIMULATE_COMPONENT_TEMPLATES_SUBSTITUTIONS = def(8_743_0_00);
    public static final TransportVersion ML_INFERENCE_IBM_WATSONX_EMBEDDINGS_ADDED = def(8_744_0_00);
    public static final TransportVersion BULK_INCREMENTAL_STATE = def(8_745_0_00);
    public static final TransportVersion FAILURE_STORE_STATUS_IN_INDEX_RESPONSE = def(8_746_0_00);
    public static final TransportVersion ESQL_AGGREGATION_OPERATOR_STATUS_FINISH_NANOS = def(8_747_0_00);
    public static final TransportVersion ML_TELEMETRY_MEMORY_ADDED = def(8_748_0_00);
    public static final TransportVersion ILM_ADD_SEARCHABLE_SNAPSHOT_TOTAL_SHARDS_PER_NODE = def(8_749_0_00);
    public static final TransportVersion SEMANTIC_TEXT_SEARCH_INFERENCE_ID = def(8_750_0_00);
    public static final TransportVersion ML_INFERENCE_CHUNKING_SETTINGS = def(8_751_0_00);
    public static final TransportVersion SEMANTIC_QUERY_INNER_HITS = def(8_752_0_00);
    public static final TransportVersion RETAIN_ILM_STEP_INFO = def(8_753_0_00);
    public static final TransportVersion ADD_DATA_STREAM_OPTIONS = def(8_754_0_00);
    public static final TransportVersion CCS_REMOTE_TELEMETRY_STATS = def(8_755_0_00);
    public static final TransportVersion ESQL_CCS_EXECUTION_INFO = def(8_756_0_00);
    public static final TransportVersion REGEX_AND_RANGE_INTERVAL_QUERIES = def(8_757_0_00);
    public static final TransportVersion RRF_QUERY_REWRITE = def(8_758_0_00);
    public static final TransportVersion SEARCH_FAILURE_STATS = def(8_759_0_00);
    public static final TransportVersion INGEST_GEO_DATABASE_PROVIDERS = def(8_760_0_00);
    public static final TransportVersion DATE_TIME_DOC_VALUES_LOCALES = def(8_761_0_00);
    public static final TransportVersion FAST_REFRESH_RCO = def(8_762_0_00);
    public static final TransportVersion TEXT_SIMILARITY_RERANKER_QUERY_REWRITE = def(8_763_0_00);
    public static final TransportVersion SIMULATE_INDEX_TEMPLATES_SUBSTITUTIONS = def(8_764_0_00);
    public static final TransportVersion RETRIEVERS_TELEMETRY_ADDED = def(8_765_0_00);
    public static final TransportVersion ESQL_CACHED_STRING_SERIALIZATION = def(8_766_0_00);
    public static final TransportVersion CHUNK_SENTENCE_OVERLAP_SETTING_ADDED = def(8_767_0_00);
    public static final TransportVersion OPT_IN_ESQL_CCS_EXECUTION_INFO = def(8_768_0_00);
    public static final TransportVersion QUERY_RULE_TEST_API = def(8_769_0_00);
    public static final TransportVersion ESQL_PER_AGGREGATE_FILTER = def(8_770_0_00);
    public static final TransportVersion ML_INFERENCE_ATTACH_TO_EXISTSING_DEPLOYMENT = def(8_771_0_00);
    public static final TransportVersion CONVERT_FAILURE_STORE_OPTIONS_TO_SELECTOR_OPTIONS_INTERNALLY = def(8_772_0_00);
    public static final TransportVersion INFERENCE_DONT_PERSIST_ON_READ_BACKPORT_8_16 = def(8_772_0_01);
    public static final TransportVersion ADD_COMPATIBILITY_VERSIONS_TO_NODE_INFO_BACKPORT_8_16 = def(8_772_0_02);
    public static final TransportVersion SKIP_INNER_HITS_SEARCH_SOURCE_BACKPORT_8_16 = def(8_772_0_03);
    public static final TransportVersion QUERY_RULES_LIST_INCLUDES_TYPES_BACKPORT_8_16 = def(8_772_0_04);
    public static final TransportVersion INITIAL_ELASTICSEARCH_8_16_5 = def(8_772_0_05);

    /*
     * STOP! READ THIS FIRST! No, really,
     *        ____ _____ ___  ____  _        ____  _____    _    ____    _____ _   _ ___ ____    _____ ___ ____  ____ _____ _
     *       / ___|_   _/ _ \|  _ \| |      |  _ \| ____|  / \  |  _ \  |_   _| | | |_ _/ ___|  |  ___|_ _|  _ \/ ___|_   _| |
     *       \___ \ | || | | | |_) | |      | |_) |  _|   / _ \ | | | |   | | | |_| || |\___ \  | |_   | || |_) \___ \ | | | |
     *        ___) || || |_| |  __/|_|      |  _ <| |___ / ___ \| |_| |   | | |  _  || | ___) | |  _|  | ||  _ < ___) || | |_|
     *       |____/ |_| \___/|_|   (_)      |_| \_\_____/_/   \_\____/    |_| |_| |_|___|____/  |_|   |___|_| \_\____/ |_| (_)
     *
     * A new transport version should be added EVERY TIME a change is made to the serialization protocol of one or more classes. Each
     * transport version should only be used in a single merged commit (apart from the BwC versions copied from o.e.Version, ≤V_8_8_1).
     *
     * ADDING A TRANSPORT VERSION
     * To add a new transport version, add a new constant at the bottom of the list, above this comment. Don't add other lines,
     * comments, etc. The version id has the following layout:
     *
     * M_NNN_S_PP
     *
     * M - The major version of Elasticsearch
     * NNN - The server version part
     * S - The subsidiary version part. It should always be 0 here, it is only used in subsidiary repositories.
     * PP - The patch version part
     *
     * To determine the id of the next TransportVersion constant, do the following:
     * - Use the same major version, unless bumping majors
     * - Bump the server version part by 1, unless creating a patch version
     * - Leave the subsidiary part as 0
     * - Bump the patch part if creating a patch version
     *
     * If a patch version is created, it should be placed sorted among the other existing constants.
     *
     * REVERTING A TRANSPORT VERSION
     *
     * If you revert a commit with a transport version change, you MUST ensure there is a NEW transport version representing the reverted
     * change. DO NOT let the transport version go backwards, it must ALWAYS be incremented.
     *
     * DETERMINING TRANSPORT VERSIONS FROM GIT HISTORY
     *
     * If your git checkout has the expected minor-version-numbered branches and the expected release-version tags then you can find the
     * transport versions known by a particular release ...
     *
     *     git show v8.11.0:server/src/main/java/org/elasticsearch/TransportVersions.java | grep '= def'
     *
     * ... or by a particular branch ...
     *
     *     git show 8.11:server/src/main/java/org/elasticsearch/TransportVersions.java | grep '= def'
     *
     * ... and you can see which versions were added in between two versions too ...
     *
     *     git diff v8.11.0..main -- server/src/main/java/org/elasticsearch/TransportVersions.java
     *
     * In branches 8.7-8.10 see server/src/main/java/org/elasticsearch/TransportVersion.java for the equivalent definitions.
     */

    /**
     * Reference to the earliest compatible transport version to this version of the codebase.
     * This should be the transport version used by the highest minor version of the previous major.
     */
    public static final TransportVersion MINIMUM_COMPATIBLE = V_7_17_0;

    /**
     * Reference to the minimum transport version that can be used with CCS.
     * This should be the transport version used by the previous minor release.
     */
    public static final TransportVersion MINIMUM_CCS_VERSION = ESQL_ATTRIBUTE_CACHED_SERIALIZATION_8_15;

    static final NavigableMap<Integer, TransportVersion> VERSION_IDS = getAllVersionIds(TransportVersions.class);

    // the highest transport version constant defined in this file, used as a fallback for TransportVersion.current()
    static final TransportVersion LATEST_DEFINED;
    static {
        LATEST_DEFINED = VERSION_IDS.lastEntry().getValue();

        // see comment on IDS field
        // now we're registered all the transport versions, we can clear the map
        IDS = null;
    }

    public static NavigableMap<Integer, TransportVersion> getAllVersionIds(Class<?> cls) {
        Map<Integer, String> versionIdFields = new HashMap<>();
        NavigableMap<Integer, TransportVersion> builder = new TreeMap<>();

        Set<String> ignore = Set.of("ZERO", "CURRENT", "MINIMUM_COMPATIBLE", "MINIMUM_CCS_VERSION");

        for (Field declaredField : cls.getFields()) {
            if (declaredField.getType().equals(TransportVersion.class)) {
                String fieldName = declaredField.getName();
                if (ignore.contains(fieldName)) {
                    continue;
                }

                TransportVersion version;
                try {
                    version = (TransportVersion) declaredField.get(null);
                } catch (IllegalAccessException e) {
                    throw new AssertionError(e);
                }
                builder.put(version.id(), version);

                if (Assertions.ENABLED) {
                    // check the version number is unique
                    var sameVersionNumber = versionIdFields.put(version.id(), fieldName);
                    assert sameVersionNumber == null
                        : "Versions ["
                            + sameVersionNumber
                            + "] and ["
                            + fieldName
                            + "] have the same version number ["
                            + version.id()
                            + "]. Each TransportVersion should have a different version number";
                }
            }
        }

        return Collections.unmodifiableNavigableMap(builder);
    }

    static Collection<TransportVersion> getAllVersions() {
        return VERSION_IDS.values();
    }

    static final IntFunction<String> VERSION_LOOKUP = ReleaseVersions.generateVersionsLookup(TransportVersions.class, LATEST_DEFINED.id());

    // no instance
    private TransportVersions() {}
}
